@@ -26,14 +26,14 @@
 #include "config.h"
 #endif
 
+#include <stdio.h>
 #include <stdint.h>
 #include <stdbool.h>
 #include <string.h>
+
 #include <glib.h>
 
-#include <stdio.h>
-
-#include <bluetooth/bluetooth.h>
+#include "lib/bluetooth.h"
 
 #include "btio/btio.h"
 #include "src/log.h"
@@ -115,6 +115,9 @@ GAttrib *g_attrib_new(GIOChannel *io, guint16 mtu)
 	if (!attr->att)
 		goto fail;
 
+	bt_att_set_close_on_unref(attr->att, true);
+	g_io_channel_set_close_on_unref(io, FALSE);
+
 	if (!bt_att_set_mtu(attr->att, mtu))
 		goto fail;
 
